import os
import json
import requests
from django.conf import settings
from slth.models import User
from slth import APPLICATON


def authenticate(code):
    for provider in APPLICATON.get('oauth', {}).values():
        client_secret = provider['client_secret']
        if client_secret.startswith('$'):
            client_secret = os.environ[client_secret[1:]]
        print(client_secret)
        redirect_uri = "{}{}".format(settings.SITE_URL, provider['redirect_uri'])
        access_token_request_data = dict(
            grant_type='authorization_code', code=code, redirect_uri=redirect_uri,
            client_id=provider['client_id'], client_secret=client_secret
        )
        response = requests.post(provider['access_token_url'], data=access_token_request_data, verify=False)
        if response.status_code == 200:
            data = json.loads(response.text)
            headers = {
                'Authorization': 'Bearer {}'.format(data.get('access_token')),
                'x-api-key': client_secret
            }
            if provider.get('user_data_method', 'GET').upper() == 'POST':
                response = requests.post(provider['user_data_url'], data={'scope': data.get('scope')}, headers=headers)
            else:
                response = requests.get(provider['user_data_url'], data={'scope': data.get('scope')}, headers=headers)
            if response.status_code == 200:
                data = json.loads(response.text)
                username = data[provider['user_data']['username']]
                user = User.objects.filter(username=username).first()
                if user:
                    return user
                elif provider.get('user_data').get('create'):
                    user = User.objects.create(
                        username=username,
                        email=data[provider['user_data']['email']] if provider['user_data']['email'] else ''
                    )
                    return user
        else:
            print(response.text)
    return

def providers():
    oauth = []
    for provider in APPLICATON.get('oauth', {}).values():
        redirect_uri = "{}{}".format(settings.SITE_URL, provider['redirect_uri'])
        authorize_url = '{}?response_type=code&client_id={}&redirect_uri={}'.format(
            provider['authorize_url'], provider['client_id'], redirect_uri
        )
        if provider.get('scope'):
            authorize_url = '{}&scope={}'.format(authorize_url, provider.get('scope'))
        oauth.append(dict(label=f'Entrar com {provider["name"]}', url=authorize_url))
    return oauth



import re
import os
import yaml
import warnings
from pathlib import Path
from django.apps import apps
from django.db import models
from .queryset import QuerySet
from django.db.models import manager
from .serializer import Serializer
from .factory import FormFactory
import django.db.models.options as options
from django.db.models.base import ModelBase
from django.core.exceptions import FieldDoesNotExist
from django.utils.autoreload import autoreload_started

warnings.filterwarnings('ignore', module='urllib3')

FILENAME = 'application.yml'
ENDPOINTS = {}
PROXIED_MODELS = []
APPLICATON = None

if APPLICATON is None and os.path.exists(FILENAME):
    with open(FILENAME) as file:
        content = file.read()
        for variable in re.findall(r'\$[a-zA-z0-9_]+', content):
            content = content.replace(variable, os.environ.get(variable[1:], ''))
    APPLICATON = yaml.safe_load(content).get('application')


class BaseManager(manager.BaseManager):
    def get_queryset(self):
        return super().get_queryset()

    def all(self):
        return self.get_queryset().all()

    def __call__(self, model):
        return apps.get_model(model)


class Manager(BaseManager.from_queryset(QuerySet)):
    pass


models.QuerySet = QuerySet
models.Manager = Manager
setattr(options, 'DEFAULT_NAMES', options.DEFAULT_NAMES + ('icon', 'search_fields'))



class ModelMixin(object):

    @classmethod
    def get_field(cls, lookup):
        model = cls
        attrs = lookup.split('__')
        while attrs:
            attr_name = attrs.pop(0)
            if attrs:  # go deeper
                field = model._meta.get_field(attr_name)
                model = field.related_model
            else:
                try:
                    return model._meta.get_field(attr_name)
                except FieldDoesNotExist:
                    pass
        return None

    def getroles(self, username_lookup='username'):
        roles = getattr(self, '_roles', None)
        if roles is None:
            obj = self
            for attr_name in username_lookup.split('__'):
                obj = getattr(obj, attr_name)
            roles = apps.get_model('api.role').objects.filter(username=obj)
            setattr(self, '_roles', roles)
        return roles

    def getuser(self, username_lookup):
        obj = self
        for attr_name in username_lookup.split('__'):
            obj = getattr(obj, attr_name)
        return apps.get_model('auth.user').objects.get(username=obj)
    
    def serializer(self) -> Serializer:
        return Serializer(self)
    
    def formfactory(self) -> FormFactory:
        return FormFactory(self)


___new___ = ModelBase.__new__


def __new__(mcs, name, bases, attrs, **kwargs):
    if attrs['__module__'] != '__fake__':
        # See .db.models.Manager
        if 'objects' in attrs and isinstance(attrs['objects'], QuerySet):
            queryset_class = attrs['objects']
            attrs.update(objects=BaseManager.from_queryset(type(queryset_class))())
        # Defining the objects Manager using .db.models.QuerySet
        if 'objects' not in attrs and not all(['objects' in dir(cls) for cls in bases]):
            attrs.update(objects=BaseManager.from_queryset(QuerySet)())

    if ModelMixin not in bases:
        bases = bases + (ModelMixin, )
    cls = ___new___(mcs, name, bases, attrs, **kwargs)
    if cls._meta.proxy_for_model:
        PROXIED_MODELS.append(cls._meta.proxy_for_model)
    return cls


ModelBase.__new__ = __new__


def api_watchdog(sender, **kwargs):
    sender.extra_files.add(Path('application.yml'))

autoreload_started.connect(api_watchdog)


def meta(verbose_name=None):
    def decorate(function):
        function.verbose_name = verbose_name
        return function
    return decorate



import slth
from django.apps import apps
from django.db import models
from django.db.models import manager, Q, CharField, CASCADE
from django.db.models.aggregates import Sum, Avg
from django.db.models.base import ModelBase
from .statistics import Statistics
from functools import reduce
from datetime import datetime
import operator
import json
from datetime import date, datetime, timedelta
from decimal import Decimal
from django.template.loader import render_to_string
from django.db import models
from .serializer import Serializer
from .exceptions import JsonResponseException
from .utils import absolute_url, append_url, build_url


class QuerySet(models.QuerySet):

    def __init__(self, *args, **kwargs):
        self.metadata = {}
        self.request = None
        super().__init__(*args, **kwargs)

    def total(self, x=None):
        return self.values_list(x).order_by(x).distinct().count() if x else self.count()

    def counter(self, x=None, y=None, title=None, chart=None):
        return Statistics(self, title=title, chart=chart).count(x=x, y=y) if x else super().count()

    def sum(self, z, x=None, y=None, title=None, chart=None):
        return Statistics(self, title=title, chart=chart).sum(z, x=x, y=y) if x else super().aggregate(sum=Sum(z))['sum'] or 0

    def _clone(self):
        qs = super()._clone()
        qs.request = self.request
        for k, v in self.metadata.items():
            v = self.metadata[k]
            if isinstance(v, list):
                qs.metadata[k] = list(v)
            elif isinstance(v, dict):
                qs.metadata[k] = dict(v)
            else:
                qs.metadata[k] = v
        return qs

    def fields(self, *names):
        self.metadata['fields'] = ('id',) + names if 'id' not in names else names
        if [name for name in names if not isinstance(name, str)]:
            return self.rows()
        return self
    
    def serializer(self, serializer):
        self.metadata['serializer'] = serializer
        return self

    def search(self, *names):
        if 'search' not in self.metadata:
            self.metadata['search'] = [name if '__' in name else '{}__icontains'.format(name) for name in names]
        return self

    def filters(self, *names):
        self.metadata['filters'] = names
        return self

    def actions(self, *names, **replacement):
        if 'actions' not in self.metadata:
            self.metadata['actions'] = []
        self.metadata['actions'].extend(names)
        for k, v in replacement.items():
            if k in self.metadata['actions']:
                self.metadata['actions'].remove(k)
            self.metadata['actions'].append(v)
        return self

    def subsets(self, *names):
        self.metadata['subsets'] = names
        return self
    
    def renderer(self, name):
        self.metadata['renderer'] = name
        return self
    
    def timeline(self):
        return self.renderer('timeline')
    
    def accordion(self):
        return self.renderer('accordion')
    
    def cards(self):
        return self.renderer('cards')
    
    def rows(self):
        return self.renderer('rows')
    
    def bi(self, *names):
        self.metadata['bi'] = []
        for name in names:
            self.metadata['bi'].append((name,) if isinstance(name, str) else name)
        return self

    def settitle(self, name=None):
        self.metadata['title'] = name or str(self.model._meta.verbose_name_plural)
        return self
    
    def attrname(self, name):
        self.metadata['attrname'] = name
        return self
    
    def reloadable(self):
        self.metadata['reloadable'] = True
        return self
    
    def calendar(self, name):
        self.metadata['calendar'] = name
        return self
    
    def related_values(self, **kwargs):
        self.metadata['relations'] = {
            k: v if isinstance(v, int) else v.pk for k, v in kwargs.items()
        }
        return self
    
    def limit(self, limit, *limits):
        self.metadata['limit'] = limit
        self.metadata['page_sizes'] = (limit,) + limits
        return self
    
    def lookup(self, role_name=None, **lookups):
        if 'lookups' not in self.metadata:
            self.metadata['lookups'] = {}
        self.metadata['lookups'][role_name] = lookups
        return self

    def apply_lookups(self, user):
        from . import permissions
        lookups = self.metadata.get('lookups')
        if lookups:
            return permissions.apply_lookups(self, lookups, user)
        return self

    def contextualize(self, request):
        pk = request and request.GET.get('id')
        choices_field_name = request and request.GET.get('choices')
        self.request = request
        qs = self._clone()
        if pk:
            if self.metadata.get('attrname') is None or request.GET.get('only') == self.attrname:
                qs = qs.filter(pk=pk)
    
        if self.request and 'action' in self.request.GET:
            cls = slth.ENDPOINTS[self.request.GET.get('action')]
            actions = self.metadata.get('actions', ())
            if cls.get_api_name() in actions:
                source = self.model.objects.get(pk=pk) if pk else self
                endpoint = cls.instantiate(self.request, source)
                if endpoint.check_permission():
                    raise JsonResponseException(endpoint.serialize())
                raise Exception()

        filters = qs.metadata.get('filters', ())
        for lookup in filters:
            if lookup.endswith('userrole'):
                from api.models import Role
                rolename = qs.parameter('userrole')
                if rolename:
                    qs = qs.filter(
                        **{'{}__in'.format(lookup[0:-10]):
                        Role.objects.filter(name=rolename).values_list('username', flat=True)}
                    )
            elif request and lookup in request.GET and lookup != choices_field_name:
                qs = qs.apply_filter(lookup, request.GET[lookup])
        search = qs.metadata.get('search', ())
        if request and search and 'q' in request.GET:
            qs = qs.apply_search(request.GET['q'], search)

        if choices_field_name:
            field = qs.model.get_field(choices_field_name)
            choices = field.related_model.objects.filter(pk__in=qs.values_list(choices_field_name, flat=True))
            term = self.request.GET.get('term')
            if term:
                choices = choices.apply_search(term)
            raise JsonResponseException([dict(id=obj.id, value=str(obj).strip()) for obj in choices[0:20]])


        return qs

    def apply_filter(self, lookup, value):
        booleans = dict(true=True, false=False, null=None)
        if len(value) == 10 and '-' in value:
            value = datetime.strptime(value, '%Y-%m-%d');
        if value in booleans:
            value = booleans[value]
        return self.filter(**{lookup: value}) if value != '' else self

    def apply_search(self, term, lookups=None):
        if lookups is None:
            search_fields = getattr(self.model._meta, 'search_fields', None)
            if search_fields is None:
                search_fields = [field.name for field in self.model._meta.get_fields() if isinstance(field, CharField)]
            lookups = [f'{name}__icontains' for name in search_fields]
        else:
            lookups = [name if name.endswith('__icontains') else f'{name}__icontains' for name in lookups]
        if lookups:
            terms = term.split(' ') if term else []
            conditions = []
            for term in terms:
                queries = [
                    Q(**{lookup: term})
                    for lookup in lookups
                ]
                conditions.append(reduce(operator.or_, queries))
            return self.filter(reduce(operator.and_, conditions)) if conditions else self
        return self
    
    def parameter(self, name, default=None):
        return self.request.GET.get(name, default) if self.request else default
    
    def serialize(self, debug=False, forward_exception=False):
        try:
            return self.to_dict(debug=debug)
        except JsonResponseException as e:
            if debug:
                print(json.dumps(e.data, indent=2, ensure_ascii=False))
            return e.data
    
    def to_dict(self, debug=False):
        title = self.metadata.get('title', str(self.model._meta.verbose_name_plural))
        title = title.title() if title and title.islower() else title
        attrname = self.metadata.get('attrname')
        relations = self.metadata.get('relations')
        renderer = self.metadata.get('renderer')
        reloadable = self.metadata.get('reloadable')
        bi = self.metadata.get('bi', [])
        subset = None
        actions = []
        filters = []
        search = []
        subsets = []
        aggregations = []
        template = None
        calendar = None

        instance_actions = []
        queryset_actions = []

        base_url = append_url(build_url(self.request), 'only={}'.format(attrname) if attrname else '')
        if self.request.GET.urlencode():
            base_url = append_url(base_url, self.request.GET.urlencode())

        for qualified_name in self.metadata.get('actions', ()):
            cls = slth.ENDPOINTS[qualified_name]
            if cls.has_args():
                instance_actions.append(cls)
            else:
                queryset_actions.append(cls)

        for cls in queryset_actions:
            if cls.instantiate(self.request, self).check_permission():
                action = cls.get_api_metadata(self.request, base_url)
                action['name'] = action['name'].replace(" {}".format(self.model._meta.verbose_name.title()), "")
                if relations and cls.get_metadata('modal'):
                        params = '&'.join(f'{k}={v}' for k, v in relations.items())
                        action['url'] = append_url(action['url'], params)
                actions.append(action)

        subset = self.parameter('subset')
        qs = self.filter()
        qs = qs.order_by('id') if not qs.ordered else qs
        if subset:
            qs = getattr(qs, subset)()
        if 'calendar' in qs.metadata:
            qs, calendar = qs.to_calendar(qs.metadata['calendar'])
        template = self.metadata.get('template')

        for name in qs.metadata.get('search', ()):
            search.append(name)

        for lookup in qs.metadata.get('filters', ()):
            if lookup.endswith('userrole'):
                field = self.role_filter_field()
            else:
                field = self.filter_field(lookup, base_url)
            if field:
                filters.append(field)

        total = qs.count()
        for name in self.metadata.get('subsets', ()):
            attr = getattr(self, name)
            label = getattr(attr, 'verbose_name', name.title())
            subsets.append(dict(name=name, label=label, count=attr().count()))
        if subsets:
            subsets.insert(0, dict(name=None, label='Tudo', count=total))

        for name in self.metadata.get('aggregations', ()):
            api_name = name[4:] if name.startswith('get_') else name
            aggregation = dict(name=api_name, label=api_name, value=getattr(self, name)())
            if isinstance(aggregation['value'], Decimal):
                aggregation['value'] = str(aggregation['value']).replace('.', ',')
            aggregations.append(aggregation)

        data = dict(type='queryset', title=title, key=attrname, url=base_url, total=total, count=total, icon=None, actions=actions, filters=filters, search=search, q=self.request.GET.get('q'))
        if renderer:
            data.update(renderer=renderer)
        if reloadable:
            data.update(reloadable=reloadable)
        if bi:
            bi_data = []
            for names in bi:
                items = []
                for name in names:
                    attr = getattr(qs, name)
                    title = getattr(attr, 'verbose_name', name.title())
                    value = attr()
                    if isinstance(value, dict):
                        item = value
                        item['title'] = title
                    else:
                        item = dict(type='counter', title=title, value=value)
                    items.append(item)
                bi_data.append(items)
            data.update(bi=bi_data)
        else:
            objs = []
            page_size = min(int(self.parameter('page_size', self.metadata.get('limit', 20))), 1000)
            page_sizes = self.metadata.get('page_sizes', [page_size])
            pages = (total // page_size) + (1 if total % page_size else 0)
            page = min(int(self.parameter('page', 1)), pages) or 1
            start = page_size * (page - 1)
            end = start + page_size
            previous = None if page ==1 else page -1
            next = None if page == pages else page + 1
            count = qs[start:end].count()
            title_func_name = '__str__'
            serializer:Serializer = qs.metadata.get('serializer')
            if serializer is None:
                fields = qs.metadata.get('fields', [field.name for field in (qs.model._meta.fields + qs.model._meta.many_to_many)])
                if relations:
                    names = set(relations.keys())
                    fields = [field_name for field_name in fields if field_name not in names]
                list_fields = [
                    f'get_{field_name}' if hasattr(qs.model, f'get_{field_name}') else
                    (f'get_{field_name}_display' if hasattr(qs.model, f'get_{field_name}_display')
                    else field_name) for field_name in fields
                ]
                serializer = Serializer(None, self.request).fields(*list_fields)
            serializer.request = self.request
            
            for obj in qs[start:end]:
                serializer.title = getattr(obj, title_func_name)()
                serializer.obj = obj
                serialized = serializer.serialize(forward_exception=True)
                for cls in instance_actions:
                    if cls.instantiate(self.request, obj).check_permission():
                        action = cls.get_api_metadata(self.request, base_url, obj.pk)
                        action['name'] = action['name'].replace(" {}".format(self.model._meta.verbose_name), "")
                        if relations and cls.get_metadata('modal'):
                            params = '&'.join(f'{k}={v}' for k, v in relations.items())
                            action['url'] = append_url(action['url'], params)
                        serialized['actions'].append(action)
                objs.append(serialized)

            if attrname:
                data.update(attrname=attrname)
            if calendar:
                data.update(calendar=calendar)
                data['filters'].extend([
                    {'name': '{}__day'.format(calendar['field']), 'type': 'hidden', 'value': calendar['day']},
                    {'name': '{}__month'.format(calendar['field']), 'type': 'hidden', 'value': calendar['month']},
                    {'name': '{}__year'.format(calendar['field']), 'type': 'hidden', 'value': calendar['year']}
                ])
            if subsets:
                data.update(subsets=subsets, subset=subset)
            if aggregations:
                data.update(aggregations=aggregations)
            if template:
                data.update(html=render_to_string(template, data))
            pagination = dict(
                start=start+1, end=min(end, total), page=dict(
                    total=pages, previous=previous, current=page, next=next,
                    size=page_size, sizes=page_sizes
                )
            )
            data.update(count=count, data=objs, pagination=pagination)
        if debug:
            print(json.dumps(data, indent=2, ensure_ascii=False))
        return data
    
    def role_filter_field(self):
        value = self.parameter('userrole')
        choices = [{'id': '', 'text': ''}]
        choices.extend({'id': k, 'text': v} for k, v in ['adm', 'Administrador'])
        return dict(name='userrole', type='choice', value=value, label='Papel', required=False, mask=None, choices=choices)

    def filter_field(self, lookups, base_url):
        name = lookups.strip()
        suffix = None
        choices = None
        field = None
        value = None
        ignore = 'icontains', 'contains', 'gt', 'gte', 'lt', 'lte', 'id', 'year', 'month'
        tmp = []
        for lookup in name.split('__'):
            if lookup in ignore:
                suffix = lookup
            else:
                tmp.append(lookup)
        field = self.model.get_field('__'.join(tmp)) if tmp else None
        if field:
            if isinstance(field, models.CharField):
                field_type = 'text'
                value = self.parameter(name)
            elif isinstance(field, models.BooleanField):
                field_type = 'choice'
                value = self.parameter(name)
                if value:
                    value = value == 'true'
                choices = [dict(id='', value=''), dict(id='true', value='Sim'), dict(id='false', value='Não'), dict(id='null', value='Nulo')]
            elif isinstance(field, models.DateField):
                field_type = 'text' if suffix in ('year', 'month') else 'date'
                value = self.parameter(name)
            elif isinstance(field, models.ForeignKey):
                field_type = 'choice'
                value = self.parameter(name)
                if value:
                    value = dict(id=value, label=self.parameter(f'{name}__autocomplete'))
                choices = append_url(base_url, f'choices={name}')
            elif isinstance(field, models.ManyToManyField):
                field_type = 'choice'
                choices = append_url(base_url, f'choices={name}')
            if getattr(field, 'choices'):
                field_type = 'choice'
                choices = [{'id': '', 'value':''}]
                choices.extend([{'id': k, 'value': v} for k, v in field.choices])
                choices.append({'id': 'null', 'value':'Nulo'})
            data = dict(name=name, type=field_type, value=value, label=str(field.verbose_name).title(), required=False, mask=None)
            if choices:
                data.update(choices=choices)
            return data
        return None

    def to_calendar(self, attr_name):
        today = date.today()
        day = self.parameter(f'{attr_name}__day')
        month = self.parameter(f'{attr_name}__month')
        year = self.parameter(f'{attr_name}__year')
        if month and year:
            start = date(int(year), int(month), 1)
        else:
            start = self.filter(**{f'{attr_name}__month': today.month}).values_list(attr_name, flat=True).first()
            if start is None:
                start = self.order_by(attr_name).values_list(attr_name, flat=True).first() or today
            month = start.month
            year = start.year
        current = date(start.year, start.month, 1)
        qs = self.filter(**{f'{attr_name}__year': start.year, f'{attr_name}__month': start.month})
        total = {}
        for key in qs.values_list(attr_name, flat=True):
            key = key.strftime('%d/%m/%Y')
            if key not in total:
                total[key] = 0
            total[key] += 1
        if day:
            qs = qs.filter(**{f'{attr_name}__day': day})
        next = current + timedelta(days=31)
        previous = current + timedelta(days=-1)
        return qs, dict(
            field=attr_name, total=total,
            day=day, month=month, year=year, next=dict(month=next.month, year=next.year),
            previous=dict(month=previous.month, year=previous.year)
        )


from datetime import date, timedelta, datetime
from django.template.loader import render_to_string


SUCCESS = "success"
PRIMARY = "primary"
WARNING = "warning"
DANGER = "danger"

class Text(dict):
    def __init__(self, text, color="inherite", size="inherite", bold=False, italic=False):
        self["type"] = "text"
        self["text"] = text
        self["color"] = color
        self["bold"] = bold
        self["italic"] = italic

class Image(dict):
    def __init__(self, src, width=None, height=None, round=False, placeholder=None):
        if width is None and height is None:
            width = 100
            height = 100
        if width and not height:
            height = width
        if height and not width:
            width = height
        self["type"] = "image"
        self["src"] = src
        self["width"] = width
        self["height"] = height
        self["round"] = round
        self["placeholder"] = placeholder


class FileLink(dict):
    def __init__(self, url, modal=False, icon=None):
        self["type"] = "filelink"
        self["url"] = url
        self["modal"] = modal
        self["icon"] = icon


class FileViewer(dict):
    def __init__(self, url):
        self["type"] = "filepreview"
        self["url"] = url


class QrCode(dict):
    def __init__(self, text):
        self["type"] = "qrcode"
        self["text"] = text


class Progress(dict):
    def __init__(self, value, style="primary"):
        self["type"] = "progress"
        self["value"] = int(value or 0)
        self["style"] = style


class Status(dict):
    def __init__(self, style, label):
        self["type"] = "status"
        self["style"] = style
        self["label"] = str(label)


class Badge(dict):
    def __init__(self, color, label, icon=None):
        self["type"] = "badge"
        self["color"] = color
        self["label"] = str(label)
        self["icon"] = icon


class Shell(dict):
    def __init__(self, output):
        self["type"] = "shell"
        self["output"] = output


class Indicators(dict):
    def __init__(self, title):
        self["type"] = "indicators"
        self["title"] = title
        self["items"] = []
        self["actions"] = []

    def append(self, name, value):
        self["items"].append(dict(name=str(name), value=value))

    def action(self, label, url, modal=False):
        self["actions"].append(dict(label=str(label), url=url, modal=modal))


class Boxes(dict):
    def __init__(self, title):
        self["type"] = "boxes"
        self["title"] = str(title)
        self["items"] = []

    def append(self, icon, label, url, style=None):
        self["items"].append(dict(icon=icon, label=str(label), url=url, style=style))


class Info(dict):
    def __init__(self, title, message):
        self["type"] = "info"
        self["title"] = title
        self["message"] = message
        self["actions"] = []

    def action(self, label, url, modal=False, icon=None):
        self["actions"].append(dict(label=str(label), url=url, modal=modal, icon=icon))


class Warning(dict):
    def __init__(self, title, message):
        self["type"] = "warning"
        self["title"] = title
        self["message"] = message
        self["actions"] = []

    def action(self, label, url, modal=False, icon=None):
        self["actions"].append(dict(label=str(label), url=url, modal=modal, icon=icon))


class Table(dict):
    def __init__(self, title, subset=None, pagination=None):
        self["type"] = "table"
        self["title"] = title
        self["actions"] = []
        self["subsets"] = []
        self["subset"] = subset
        self["filters"] = []
        self["flags"] = []
        self["rows"] = []
        self["pagination"] = {}

    def add_subset(self, name, label, count):
        self["subsets"].append(dict(name=name, label=label, count=count))

    def add_action(self, name, label, icon=None, batch=True):
        self["actions"].append(dict(name=name, label=label, icon=icon, batch=batch))

    def add_flag(self, name, label, checked=False):
        self["flags"].append(dict(name=name, label=label, checked=checked))

    def add_filter(self, ftype, name, label, value, choices=None):
        self["filters"].append(
            dict(type=ftype, name=name, label=label, value=value, choices=choices)
        )

    def pagination(self, size, page, total, sizes):
        self["pagination"].update(size=size, page=page, total=total, sizes=sizes)

    def add_row(self, row):
        self["rows"].append(row)

    def row(self, value=None, checkable=False, deleted=False):
        self["rows"].append(
            [dict(name="#", value=value, checkable=checkable, deleted=deleted)]
        )

    def cell(self, name, value, style=None, url=None, actions=None):
        self["rows"][-1].append(
            dict(name=name, value=value, style=style, url=url, actions=actions)
        )


class TemplateContent(dict):
    def __init__(self, name, context):
        self["type"] = "html"
        self["content"] = render_to_string(name, context)


class Banner(dict):
    def __init__(self, src):
        self["type"] = "banner"
        self["src"] = src


class Map(dict):
    def __init__(self, latitude, longitude, width="100%", height=400):
        self["type"] = "map"
        self["latitude"] = str(latitude)
        self["longitude"] = str(longitude)
        self["width"] = width
        self["height"] = height


class Steps(dict):
    def __init__(self, icon=None):
        self["type"] = "steps"
        self["icon"] = icon
        self["steps"] = []

    def append(self, name, done):
        number = len(self["steps"]) + 1
        self["steps"].append(dict(number=number, name=name, done=bool(done)))


class WebConf(dict):
    def __init__(self, caller, receiver):
        self["type"] = "webconf"
        self["caller"] = caller
        self["receiver"] = receiver


class Navbar(dict):
    def __init__(self, title, subtitle=None, logo=None, user=None):
        self["type"] = "navbar"
        self["title"] = title
        self["subtitle"] = subtitle
        self["logo"] = logo
        self["user"] = user
        self["usermenu"] = []
        self["adder"] = []
        self["tools"] = []
        self["settings"] = []
        self["actions"] = []
        self["toolbar"] = []

    def add_action(self, entrypoint, name, url, modal=True, icon=None):
        self[entrypoint].append(dict(name=name, url=url, modal=modal, icon=icon))


class Menu(dict):
    def __init__(self, items, user=None, image=None):
        self["type"] = "menu"
        self["items"] = items
        self["user"] = user
        self["image"] = image


class Footer(dict):
    def __init__(self, version):
        self["type"] = "navbar"
        self["version"] = version


class Application(dict):
    def __init__(self, icon=None, navbar=None, menu=None, footer=None, oauth=(), sponsors=()):
        self["type"] = "application"
        self["icon"] = icon
        self["navbar"] = navbar
        self["menu"] = menu
        self["footer"] = footer
        self["oauth"] = oauth
        self["sponsors"] = sponsors


class Response(dict):

    def __init__(self, message=None, redirect=None, store=None, **kwargs):
        self["type"] = "response"
        self["message"] = message
        self["redirect"] = redirect or '..'
        self["store"] = store or {}
        self.update(**kwargs)


class IconSet(dict):
    def __init__(self):
        self["type"] = "iconset"


class Grid(dict):
    def __init__(self):
        self["type"] = "grid"
        self["items"] = []

    def append(self, component):
        self["items"].append(component)


class Scheduler(dict):
    INTERVALS = {1: ["00"], 2: ["00", "30"]}

    def __init__(
        self,
        start_time=7,
        end_time=20,
        chucks=2,
        start_day=None,
        days=14,
        single_selection=False,
        input_name="schedule",
        readonly=False,
    ):
        self["type"] = "scheduler"
        self["single_selection"] = single_selection
        self["input_name"] = input_name
        self["readonly"] = readonly
        self.end_day = start_day or datetime.now()
        self.end_day = datetime(self.end_day.year, self.end_day.month, self.end_day.day)
        self.times = []
        for hour in range(start_time, end_time + 1):
            for minute in Scheduler.INTERVALS[chucks]:
                self.times.append("{}:{}".format(str(hour).rjust(2, "0"), minute))
        self.days = []
        for n in range(0, days):
            self.days.append(self.end_day.strftime("%d/%m/%Y"))
            self.end_day = self.end_day + timedelta(days=1)
        self.end_day = datetime(self.end_day.year, self.end_day.month, self.end_day.day, 23, 59, 59)
        self["slots"] = {}
        for day in self.days:
            self["slots"][day] = {k: None for k in self.times}

        self["matrix"] = []
        row = [dict(text='', icon=None, color=None)]
        for day in self.days:
            row.append(dict(text=day, icon=None, color=None))
        self["matrix"].append(row)
        for time in self.times:
            row = [dict(text=time, icon=None, color=None)]
            for day in self.days:
                row.append(self["slots"][day][time])
            self["matrix"].append(row)

    def append(self, date_time, text, icon='check'):
        day = date_time.strftime("%d/%m/%Y")
        time = date_time.strftime("%H:%M")
        value = dict(text=text, icon=icon)
        self["slots"][day][time] = value
        j = self.days.index(day) + 1 if day in self.days else -1
        x = self.times.index(time) + 1 if time in self.times else -1
        self["matrix"][x][j] = value

        


import os
from datetime import date, timedelta
from unittest import skipIf
from django.apps import apps
from django.core.cache import cache
from django.core.management import call_command
from .selenium import SeleniumTestCase
from django.contrib.auth.models import User


# @skipIf(os.environ.get("FRONTEND_PROJECT_DIR") is None, "FRONTEND_PROJECT_DIR is not set")
class IntegrationTest(SeleniumTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def test(self):
        if self.step("1"):
            self.create_superuser("admin@tavos.com", "123")
        if self.step("2"):
            self.open("/admin/")
            self.wait(3)
            self.open('/')


import base64
import json

import requests
from django.contrib.auth.models import User
from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from .selenium import SeleniumTestCase


class HttpData:
    def __init__(self, data=None):
        self._mutable = False
        self.data = data or {}
        
    def get(self, k, default=None):
        v = self.data.get(k)
        if isinstance(v, list):
            return v[0] if v else default
        return default if v is None else v

    def getlist(self, k):
        return self.data.get(k) or []
    
    def __contains__(self, item):
        return item in self.data
    
    def __getitem__(self, item):
        return self.data[item]
    
    def clear(self):
        self.data.clear()

class HttpRequest:
    def __init__(self, querystring=None, data=None):
        self.path = ''
        self.scheme = 'http'
        self.method = 'POST' if data else 'GET'
        self.body = None
        self.GET = HttpData()
        self.POST = HttpData(data)
        self.META = {
            'CONTENT_TYPE': '_application/json'
        }
        self.user = User.objects.filter(is_superuser=True).first() or User()
        self.FILES = None

        if querystring:
            for parameter in querystring[1:].split('&'):
                name, value = parameter.split('=')
                if name not in self.GET.data:
                    self.GET.data[name] = []
                self.GET.data[name].append(value)

    def get_host(self):
        return 'localhost'
    
    def get_port(self):
        return '8000'

class ServerTestCase(StaticLiveServerTestCase):

    def __init__(self, *args, **kwargs):
        self.authorization = None
        self.debug = False
        super().__init__(*args, **kwargs)

    def get_headers(self, ):
        if self.authorization:
            headers = {}
            headers['Authorization']= self.authorization
            return headers
        return None

    def log(self, method, url, data, response):
        if self.debug:
            print('{}: {}'.format(method, url))
            try:
                if data:
                    print('Input:\n{}'.format(json.dumps(data, indent=2, ensure_ascii=False)))
                print('Output:\n{}'.format(json.dumps(response.json(), indent=2, ensure_ascii=False)))
            except Exception:
                print(response.text)
                import traceback
                print(traceback.format_exc())
                print(data)
            print('\n')

    def url(self, path):
        if path.startswith('/o/'):
            return '{}{}'.format(self.live_server_url, path)
        return '{}{}'.format(self.live_server_url, path)

    def get(self, path, data=None, status_code=200):
        url = self.url(path)
        response = requests.get(url, data=data, headers=self.get_headers())
        self.log('GET', url, data, response)
        self.assertEquals(response.status_code, status_code)
        return response.json()

    def post(self, path, data=None, json=None, status_code=200):
        url = self.url(path)
        response = requests.post(url, data=data, json=json, headers=self.get_headers())
        self.log('POST', url, data or json, response)
        self.assertEquals(response.status_code, status_code)
        return response.json()

    def put(self, path, data=None, status_code=200):
        url = self.url(path)
        response = requests.put(url, data=data, headers=self.get_headers())
        self.log('PUT', url, data, response)
        self.assertEquals(response.status_code, status_code)
        return response.json()

    def delete(self, path, data=None, status_code=200):
        url = self.url(path)
        response = requests.delete(url, data=data, headers=self.get_headers())
        self.log('DELETE', url, data, response)
        self.assertEquals(response.status_code, status_code)
        return response.json()

    def login(self, username, password):
        self.authorization = 'Basic {}'.format(
            base64.b64encode('{}:{}'.format(username, password).encode()).decode()
        )

    def authorize(self, access_token, token_type='Bearer'):
        self.authorization = '{} {}'.format(token_type, access_token)

    def logout(self):
        self.authorization = None

    def assert_model_count(self, model, n):
        self.assertEquals(self.objects(model).count(), n)

    def objects(self, model):
        return apps.get_model(model).objects
    
    def print(self, data):
        print(json.dumps(data, indent=2, ensure_ascii=False))


    @staticmethod
    def create_user(username, password, is_superuser=False):
        user = User.objects.create(username=username)
        user.set_password(password)
        user.is_superuser = is_superuser
        user.save()
        return user



import os
import signal
import warnings
from subprocess import DEVNULL, check_call

from django.apps import apps
from django.conf import settings
from django.contrib.staticfiles.handlers import StaticFilesHandler
from django.contrib.staticfiles.testing import LiveServerTestCase
from django.core.cache import cache
from django.core.management import call_command
from django.core.servers.basehttp import WSGIServer
from django.db import connection
from django.http import HttpResponse
from django.core.management import call_command
from django.contrib.auth.models import User

from .browser import Browser

WSGIServer.handle_error = lambda *args, **kwargs: None


class ContextManager:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class TestStaticFilesHandler(StaticFilesHandler):
    def _middleware_chain(self, request):
        return HttpResponse()


class SeleniumTestCase(LiveServerTestCase):
    SAVE_STEP = cache.get("SAVE_STEP")
    FROM_STEP = cache.get("FROM_STEP")
    NOBUILD = cache.get("NOBUILD")
    HEADLESS = cache.get("HEADLESS", True)
    RESTORE = cache.get("RESTORE")
    LOG_ACTION = cache.get("LOG_ACTION", False)

    FRONTEND_HOST_URL = None
    FRONTEND_PROJECT_DIR = None
    FRONTEND_PROCESS_PID = None

    static_handler = TestStaticFilesHandler

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.login_count = 0
        self.current_username = None
        warnings.filterwarnings("ignore")

        self._url = "/app/login/"
        self._execute = 1
        self._step = 0
        self._adverb = 0

    @classmethod
    def setUpClass(cls):
        call_command("sync")
        cls.host = os.environ.get('BACKEND_HOST', '127.0.0.1')
        cls.port = int(os.environ.get('BACKEND_PORT', '8000'))
        super().setUpClass()
        print(cls.live_server_url)
        cache.clear()
        # url = "http://{}:{}".format(
        #     os.environ.get('FRONTEND_HOST', '127.0.0.1'),
        #     os.environ.get('FRONTEND_PORT', '5173')
        # )
        cls.browser = Browser(cls.live_server_url, slowly=False, headless=SeleniumTestCase.HEADLESS)
        for app_label in settings.INSTALLED_APPS:
            app_module = __import__(app_label)
            app_dir = os.path.dirname(app_module.__file__)
            fixture_path = os.path.join(app_dir, "fixtures", "test.json")
            if os.path.exists(fixture_path):
                call_command("loaddata", fixture_path)
        settings.DEBUG = True

    def create_superuser(self, username, password):
        if not User.objects.filter(username=username).exists():
            User.objects.create_superuser(username, password=password)

    def wait(self, seconds=1):
        self.browser.wait(seconds)

    def open(self, url="/"):
        self.browser.open(url)
        self._url = url

    def login(self, username, password):
        self.open('/app/login/')
        self.enter('Username', username)
        self.enter('Senha', password)
        self.click('Enviar')

    def logout(self, username):
        self.click(username)
        self.click('Sair')

    def slow(self, slowly=False):
        self.browser.slow(slowly)

    def reload(self):
        self.browser.reload()

    def enter(self, name, value, submit=False, count=5):
        self.browser.enter(name, value, submit, count)

    def choose(self, name, value, count=5):
        self.browser.choose(name, value, count)

    def see(self, text, flag=True, count=5):
        self.browser.see(text, flag, count)

    def look_at(self, text, count=5):
        self.browser.look_at(text, count)
        return ContextManager()

    def click(self, text):
        self.browser.click(text)

    def step(self, description=None):
        if not SeleniumTestCase.FROM_STEP and not SeleniumTestCase.SAVE_STEP:
            return True
        if self._step > 1 and self._execute:
            self.save()
        if SeleniumTestCase.RESTORE and self.step_file_exists(SeleniumTestCase.RESTORE) and self._execute:
            print(f"Creating development database from existing step {SeleniumTestCase.RESTORE}")
            self.create_dev_database(self.step_file_exists(SeleniumTestCase.RESTORE))
            self._execute = 0
            return False
        self._step += 1
        if SeleniumTestCase.FROM_STEP:
            if self._step == SeleniumTestCase.FROM_STEP:
                self.load(self._step)
                self._execute = 2
                return False
            else:
                self._execute = 2 if self._execute > 1 else 0

        if self._execute:
            description = "Executando passo {}{}".format(self._step, f" ({description})" if description else "")
            print(description)
        return self._execute

    def tearDown(self):
        self.save()
        if SeleniumTestCase.LOG_ACTION:
            SeleniumTestCase.LOG_ACTION = 2
            self.reload()
            input("Logging the actions in the terminal. Type any key to exit!\n\n")
        return super().tearDown()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

        try:
            if cls.FRONTEND_PROCESS_PID:
                os.killpg(os.getpgid(cls.FRONTEND_PROCESS_PID), signal.SIGTERM)
        except ProcessLookupError:
            pass

        cls.browser.close()

    def postgres_parameters(self):
        dbhost = settings.DATABASES["default"]["HOST"]
        dbuser = settings.DATABASES["default"]["USER"]
        dbport = settings.DATABASES["default"]["PORT"]
        dbparam = f"-U {dbuser} -h {dbhost} -p {dbport}"
        return dbparam

    def create_dev_database(self, fname=None):
        dbname = settings.DATABASES["default"]["NAME"]
        if "sqlite3" in settings.DATABASES["default"]["ENGINE"]:
            if os.path.exists("db.sqlite3"):
                os.unlink("db.sqlite3")
            os.system('sqlite3 db.sqlite3 "VACUUM;"')
            if fname:
                os.system(f"cat {fname} | sqlite3 db.sqlite3")
            else:
                os.system(f'sqlite3 {dbname} ".dump" | sqlite3 db.sqlite3')
        elif "postgresql" in settings.DATABASES["default"]["ENGINE"]:
            dbparam = self.postgres_parameters()
            dbname2 = dbname[5:]
            os.system(f"dropdb {dbparam} --if-exists {dbname2}")
            os.system(f"createdb {dbparam} {dbname2}")
            if fname:
                os.system(f"pg_dump {dbparam} --schema-only -d {dbname} | psql {dbparam} -q -d {dbname2} > /dev/null")
                os.system(f"cat {fname} | psql {dbparam} -q -d {dbname2} > /dev/null")
            else:
                os.system(f"pg_dump {dbparam} -d {dbname} | psql {dbparam} -q -d {dbname2} > /dev/null")

    def save(self):
        if not SeleniumTestCase.SAVE_STEP:
            return
        if SeleniumTestCase.RESTORE == self._step:
            print(f"Creating development database from step {self._step}")
            self.create_dev_database()
            self._execute = 0
        if self._execute:
            print(f"Saving step {self._step}")
            os.makedirs(".steps", exist_ok=True)
            dbname = settings.DATABASES["default"]["NAME"]
            fname = "{}.sql".format(os.path.join(".steps", str(self._step)))
            if "sqlite3" in settings.DATABASES["default"]["ENGINE"]:
                cmd = f'sqlite3 {dbname} ".dump" > {fname}'
                if "memory" not in cmd:
                    os.system(cmd)
            elif "postgresql" in settings.DATABASES["default"]["ENGINE"]:
                cmd = "pg_dump {} -d {} --inserts --data-only --no-owner -f {}".format(
                    self.postgres_parameters(), dbname, fname
                )
                check_call(cmd.split(), stdout=DEVNULL, stderr=DEVNULL)

    def step_file_exists(self, step):
        fname = "{}.sql".format(os.path.join(".steps", str(step)))
        return fname if os.path.exists(fname) else False

    def load(self, step):
        fname = "{}.sql".format(os.path.join(".steps", str(self._step)))
        print(f"Loading step {step}")
        dbname = settings.DATABASES["default"]["NAME"]
        if "sqlite3" in settings.DATABASES["default"]["ENGINE"]:
            cmd = f"""sqlite3 {dbname} \"PRAGMA writable_schema = 1;
                    delete from sqlite_master where type in ('table', 'index', 'trigger');
                    PRAGMA writable_schema = 0;VACUUM;PRAGMA integrity_check;\""""
            os.system(cmd)
            cmd = f"cat {fname} | sqlite3 {dbname}"
            os.system(cmd)
        else:
            cursor = connection.cursor()
            tables = [m._meta.db_table for c in apps.get_app_configs() for m in c.get_models()]
            for table in tables:
                cursor.execute(f"truncate table {table} cascade;")
            cmd = f"psql {self.postgres_parameters()} -d {dbname} --file={fname}"
            check_call(cmd.split(), stdout=DEVNULL, stderr=DEVNULL)

    def loaddata(self, fixture_path):
        call_command("loaddata", "--skip-checks", fixture_path)



import slth
from django.apps import apps
from django.conf import settings
from django.urls import path, re_path
from . import views

urlpatterns = [
    # re_path(r'^(?P<path>.*)/$', views.dispatcher),
]

for app_label in settings.INSTALLED_APPS:
    try:
        app = apps.get_app_config(app_label.split('.')[-1])
        fromlist = app.module.__package__.split('.')
        if app_label != 'slth':
            __import__('{}.{}'.format(app_label, 'endpoints'), fromlist=fromlist)
    except ImportError as e:
        if not e.name.endswith('endpoints'):
            raise e
    except BaseException as e:
        raise e


for cls in slth.ENDPOINTS.values():
    pattern = cls.get_api_url_pattern()
    urlpatterns.append(path(pattern, views.dispatcher))
    #print(f'/api/{pattern}')
#print()
